package tictactoe;

import java.io.*;
import java.util.*;
import tictactoe.*;

/**
 * AbstractBoardGame implements common methods to
 * TicTacToe and Gomoku.
 *
 * @see BoardGame
 * @see TicTacToe
 * @see Gomoku
 *
 * @author Oscar.Nierstrasz@acm.org
 * @version 2.1 1999-03-31
 */
public abstract class AbstractBoardGame
		extends Observable implements BoardGame
{
	protected int _connected = 0; // number of clients connected

	protected Player[][] _gameState;
	protected int _rows;
	protected int _cols;
	protected int _winningScore;

	protected final Player _nobody = new InactivePlayer();
	protected Player _winner = _nobody;
	protected Player[] _player;
	protected int _turn = X; // initial turn
	protected int _squaresLeft = 9;
	
	static final int X = 0;
	static final int O = 1;

	/**
	 * Generic initialization for BoardGame.
	 * The state of the game is represented as an
	 * array of Players.
	 *
	 * @param cols the number of columns
	 * @param rows the number of rows
	 * @param score the winning score (3 for TicTacToe; 5 for Gomoku)
	 * @param playerX the player whose mark is 'X'
	 * @param playerO the player whose mark is 'Y'
	 */
	protected void init(int cols, int rows, int score,
		Player playerX, Player playerO)
	{
		_rows = rows;
		_cols = cols;
		_gameState = new Player[_cols][_rows];
		for (int col=0; col <_cols; col++)
			for (int row=0; row<_rows; row++)
				try {
					this.set(col,row,_nobody);
				} catch(AssertionException err) {
					initFailed(err);
				}
		_squaresLeft = _rows * _cols;
		
		_winningScore = score;
		
		_player = new Player[2];
		_player[X] = playerX;
		_player[O] = playerO;

		try {
			assertTrue(_nobody.isNobody());
			assertTrue(playerX.mark() == 'X');
			assertTrue(playerO.mark() == 'O');
		} catch(AssertionException err) {
			initFailed(err);
		}
		
		playerX.setGame(this);
		playerO.setGame(this);
	}
	
	protected void initFailed(AssertionException err) {
		System.err.println("Fatal error -- BoardGame initialization failed!");
		err.printStackTrace();
		System.exit(1);
	}

	/**
	 * The game is ready to be played.
	 * (There are two connected clients.)
	 */
	public boolean ready() {
		return _connected >= 2;
	}

	/**
	 * A client connected to the game must join
	 * the game to know what its mark is (X or O).
	 * If two players are connected, further players
	 * may connect, but only as spectators.
	 */
	public char join() {
		char mark;
		if (_connected < 2) {
			mark = _player[_connected].mark();
		} else {
			mark = _nobody.mark(); // spectator
		}
		_connected++;
		return mark;
	}

	/**
	 * @return the number of columns of this game
	 */
	public int cols() { return _cols; }
	/**
	 * @return the number of rows
	 */
	public int rows() { return _rows; }

	/**
	 * Modifies the state of the game.  This method is only called
	 * by the BoardGame itself when a valid move has taken place.
	 *
	 * @param col the column
	 * @param row the row
	 * @param player the Player attempting the move
	 */
	protected void set(int col, int row, Player player)
		throws AssertionException
	{
		assertTrue(inRange(col, row));
		_gameState[col][row] = player;
	}

	/**
	 * Used by Runner to detect a winning run of marks.
	 * @see Runner#samePlayer
	 * @return the current mark at the given location on the Board
	 */
	public Player get(int col, int row)
		throws AssertionException
	{
		assertTrue(inRange(col, row));
		return _gameState[col][row];
	}

	public Player currentPlayer() {
		return _player[_turn];
	}

	/**
	 * Called by the current player.  Here we check if the move
	 * is valid.  If so the game state is updated and the
	 * Observers are notified after the invariant is re-established.
	 *
	 * @see StreamPlayer#move
	 * @see AppletPlayer#move
	 */
	public void move(int col, int row, Player p)
		throws AssertionException
	{
		assertTrue(this.notOver());
		assertTrue(p == currentPlayer());
		assertTrue(this.get(col, row).isNobody());
		this.set(col, row, p);
		this._squaresLeft--;
		this.swapTurn();
		this.checkWinner(col,row);
		assertTrue(this.invariant());
		setChanged();
		notifyObservers(new Move(col, row, p.mark()));
	}

	/**
	 * Attempt to move to the given square.
	 * Called by a remote client.  The Boardgame
	 * will check if move.mark corresponds to the
	 * current Player, and then call move().
	 */
	public void remoteMove(Move move)
		throws AssertionException
	{
		assertTrue(move.mark == currentPlayer().mark());
		this.move(move.col, move.row, currentPlayer());
	}

	protected void swapTurn() {
		_turn = (_turn == X) ? O : X;
	}
	
	public Player winner() {
		return _winner;
	}
	
	/**
	 * The game is not over as long as there is no winner
	 * and somebody can still make a move ...
	 */
	public boolean notOver() {
		return this.winner().isNobody()
			&& this.squaresLeft() > 0;
	}
	
	public int squaresLeft() {
		return this._squaresLeft;
	}

	/**
	 * This algorithm works both for TicTacToe and Gomoku.
	 * Create a "Runner" that starts at the current position,
	 * and runs back in forth in a given direction, returning
	 * the length of the run (i.e., number of consecutive pieces
	 * of the same Player).  If the number is big enough, the
	 * current Player wins.
	 *
	 * @param col the column of the current move
	 * @param row the row of the current move
	 */
	protected void checkWinner(int col, int row)
		throws AssertionException
	{
		Player player = this.get(col,row);
		Runner runner = new Runner(this, col, row);

		// check vertically
		if (runner.run(0,1) >= this._winningScore)
			{ this.setWinner(player); return; }
		// check horizontally
		if (runner.run(1,0) >= this._winningScore)
			{ this.setWinner(player); return; }
		// check diagonally
		if (runner.run(1,1) >= this._winningScore)
			{ this.setWinner(player); return; }
		// and check the other diagonal
		if (runner.run(1,-1) >= this._winningScore)
			{ this.setWinner(player); return; }
	}

	protected void setWinner(Player player) {
		_winner = player;
	}

	/**
	 * A plain ascii representation of the game,
	 * mainly for debugging purposes.
	 */
	public String toString() {
		StringBuffer rep = new StringBuffer();
	
		for (int row=this._rows-1; row>=0; row--) {
			rep.append(1 + row);
			if (row < 9)
				rep.append(' '); // extra space for single digit
			rep.append("  ");
			for (int col=0; col < this._cols; col++) {
				try {
					rep.append(this.get(col,row).mark());
				} catch (AssertionException err) {
					rep.append("Assertion violation: "
						+ err.getMessage());
				}
				if (col<this._cols-1) {
					rep.append(" | ");
				}
			}
			rep.append('\n');
			if (row>0) {
				rep.append("   ---");
				for (int i=1; i<this._cols; i++)
					rep.append("+---");
				rep.append("\n");
			}
		}
		rep.append(" ");
		for (int col=0; col < this._cols; col++) {
			rep.append("   ");
			rep.append((char) ('a' + col));
		}
		rep.append("\n");
		return(rep.toString());
	}

	/**
	 * @return true if the arguments identify a valid position on the board
	 */
	public boolean inRange(int col, int row) {
		return ((0<=col) && (col<this._cols)
			&& (0<=row) && (row<this._rows));
	}
	
	protected void assertTrue(boolean assertion)
		throws AssertionException
	{
		if (!assertion) {
				throw new AssertionException("assertion failed in "
					+ this.getClass().getName());
		}
	}
	
	/**
	 * Assertions to check at the start and end
	 * of state-modifying methods.
	 *
	 * These seem obvious, which is exactly why
	 * they should be checked.
	 */
	private boolean invariant() {
		return (_turn == X || _turn == O)
			&& (this.notOver() 
				|| this.winner() == _player[X]
				|| this.winner() == _player[O]
				|| this.winner().isNobody())
			&& (_squaresLeft < _rows * _cols
				// else, initially:
				|| _turn == X && this.winner().isNobody());
	}

	/**
	 * Some generic tests for both TicTacToe and Gomoku.
	 */
	public void test() {
		try {
			System.err.println("Started AbstractboardGame tests");
			assertTrue(this.invariant());

			assertTrue(this.get(0,0).isNobody());
			
			/*
			this.set(0,0,'X');
			assertTrue(this.get(0,0) == 'X');
			this.set(0,0,' ');
			assertTrue(this.get(0,0) == ' ');
			*/

			assertTrue(this.inRange(0,0));
			assertTrue(this.inRange(_cols-1,_rows-1));

			assertTrue(!this.inRange(-1,0));
			assertTrue(!this.inRange(0,-1));
			assertTrue(!this.inRange(0,_rows));
			assertTrue(!this.inRange(_cols,0));

			assertTrue(this.invariant());
			System.err.println("Passed AbstractboardGame tests");
		} catch (AssertionException err) {
			System.err.println("AbstractboardGame assertion failed!");
			err.printStackTrace();
		}
	}
}
